#!/usr/bin/env python
# coding=utf-8

'''
*Copyright 2018  Ridolfa.eu*

    *
    * Licensed under the EUPL, Version 1.1 or – as soon they    *
      will be approved by the European Commission - subsequent
      versions of the EUPL (the "Licence");
    * You may not use this work except in compliance with the
      Licence.
    * You may obtain a copy of the Licence at:
    *
    * https://joinup.ec.europa.eu/software/page/eupl
    *
    * Unless required by applicable law or agreed to in
      writing, software distributed under the Licence is
      distributed on an "AS IS" basis,
    * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
      express or implied.
    * See the Licence for the specific language governing
      permissions and limitations under the Licence.
'''

import time

# DS commands
DS_CMD_RESET            = 0xF0  # DS2482 reset
DS_CMD_WRITECONFIG      = 0xD2  # Write DS2482 config byte
DS_CMD_SETREADPOINTER   = 0xE1  # Set read pointer
DS_CMD_CHANNELSELECT    = 0xC3  # Channel select

# DS pointer codes to select a register
DS_REG_STATUS           = 0xF0  # Status Register
DS_REG_CONFIG           = 0xC3  # Configuration Register
DS_REG_DATA             = 0xE1  # Data Register
DS_REG_CHANNEL          = 0xD2  # Channel Selection Register

# 1-wire commands
OW_CMD_RESET            = 0xB4  # 1-wire reset
OW_CMD_WRITEBYTE        = 0xA5  # write byte to 1-wire
OW_CMD_READBYTE         = 0x96  # read byte from 1-wire
OW_CMD_MATCHROM         = 0x55  # specific device: listen for a command
OW_CMD_SKIPROM          = 0xCC  # all devices: listen for a command
OW_CMD_READROM          = 0x33  # read rom of a 1-wire device
OW_CMD_SEARCHROM        = 0xF0  # initiates the search rom cycle

OW_CMD_READMEMORY       = 0xF0
OW_CMD_COPYSCRATCHPAD   = 0x55
OW_CMD_SINGLEBIT        = 0x87  # 1-wire single bit
OW_CMD_TRIPLET          = 0x78  # 1-wire triplet

# Status register byte
DS_STATUS_DIR           = 0x80  # Branch Direction Taken
DS_STATUS_TSB           = 0x40  # Triplet Second Bit
DS_STATUS_SBR           = 0x20  # Single Bit Result
DS_STATUS_RST           = 0x10  # Device Reset
DS_STATUS_LL            = 0x08  # 1-Wire Logic Level
DS_STATUS_SD            = 0x04  # Short detected
DS_STATUS_PPD           = 0x02  # Presence puls detected
DS_STATUS_1WB           = 0x01  # 1-wire bus busy

# Configuration register byte                               Defaults
DS_CONFIG_i1WS          = 0x80  # invers of 1WS             1
DS_CONFIG_iSPU          = 0x40  # invers of SPU             1
DS_CONFIG_1             = 0x20  # always 1                  1
DS_CONFIG_iAPU          = 0x10  # invers of APU             1
DS_CONFIG_1WS           = 0x08  # 1-Wire Speed              0
DS_CONFIG_SPU           = 0x04  # Strong pull-up            0
DS_CONFIG_PPM           = 0x02  # Presence pulse masking    0
DS_CONFIG_APU           = 0x01  # Active pull-up            0

# Channel selection and read back code
DS_CH_CODE = [
    [0xF0,0xB8],    # 0 [Select,read] channel IO0  default at power-up
    [0xE1,0xB1],    # 1 [Select,read] channel IO1
    [0xD2,0xAA],    # 2 [Select,read] channel IO2
    [0xC3,0xA3],    # 3 [Select,read] channel IO3
    [0xB4,0x9C],    # 4 [Select,read] channel IO4
    [0xA5,0x95],    # 5 [Select,read] channel IO5
    [0x96,0x8E],    # 6 [Select,read] channel IO6
    [0x87,0x87]     # 7 [Select,read] channel IO7
    ]

OW_BUS_DELAY = 0.001   # 1-wire bus needs time to perform a command
# Raspi 2B  seems to work with time = 0 but faster CPUs like Raspi 3 need a delay.
# In particular the I2C_write_byte function is suseptible to short delay times
# and produces a pigpio error

ERROR_NO  = 1
ERROR_YES = -1

DEVDATA = [ERROR_YES,ERROR_YES,"ridolfa.eu"] # initial return data of 1-wire devices


# noinspection PyUnresolvedReferences
class DS2482():
    """
    A class to use the I2C <--> 1-wire bridges DS2482-100 and DS2482-800.
    The class supports only one device on the I2C bus.
    There are three variants:

    - Raspi, using the PIGPIO library (recommended)
        - Raspi, using the RGPIO library (work in progress)
        - ESP8266, using the I2C micropython library (work in progress)

    Subclasses for 1-wire devices are included:

    class DS18B20       1-wire thermometer
    class DS24SN        DS2401, Ds2411 Silicon Number devices
    class DS2406        dgital input/output device
    """
    def __init__(self, pi, I2C_address, DS_address):
        """

        """
        self.pi = pi
        self.dsh = self.I2C_connect(I2C_address, DS_address)
        self.channel_last = 0

    # I2C Host Operations ---------------------------------------------

    def I2C_connect(self, I2C_address, DS_address):
        """
        """
        dsh = self.pi.i2c_open(I2C_address, DS_address)
        return(dsh)



    def I2C_write_byte (self,cmd, data=None):
        """
        Writes one or two bytes to the I2C bus.

        Parameter
            cmd     DS2482 command byte
            data    data byte or 1-wire command byte
        Return
            1       write ok
            -1      write failed
       """
        #with open("i2cwrite.log",'a') as fh:
        #    x = time.strftime("%d %b %Y %H:%M:%S")
        #    print("{} {:x} {}".format(x, cmd,data),file=fh)

        if data is  None:
            if self.pi.i2c_write_byte(self.dsh,cmd) != 0:
                print("Error in function <I2C_write_byte>",hex(cmd))
                return(ERROR_YES)
        elif data is not None:
            if self.pi.i2c_write_device(self.dsh, [cmd,data]) != 0:
                print("Error in function <I2C_write_byte>",hex(cmd),hex(data))
                return(ERROR_YES)
        time.sleep(OW_BUS_DELAY)
        return(ERROR_NO)

    def I2C_readbyte(self, regpointer=ERROR_YES):
        """
        Reads a byte from the I2C bus.
        The byte read is returned from the function.
        Sets read pointer before reading if regpointer != None
        Parameter:
            regpointer
                None    read register byte from a pre-set register
                xxx     set register pointer and read register byte
            Return
                regbyte     register byte
                -1          read failed
        """

        if regpointer is not None and self.I2C_write_byte(cmd=DS_CMD_SETREADPOINTER, data=regpointer) != ERROR_NO:
            return(False)
        regbyte = self.pi.i2c_read_byte(self.dsh)
        return(regbyte)

    # Basic DS2482 commands -------------------------------------------

    def DS_reset(self):
        """
        Device Reset, Command Code:       0xF0
        Performs a global reset of device state machine logic.
        Terminates any ongoing 1-Wire communication.

        Return
            1   reset ok
            -1  reset failed
        """

        if self.I2C_write_byte(DS_CMD_RESET) != ERROR_NO: return(False)
        regbyte = self.I2C_readbyte(regpointer = DS_REG_STATUS)
        if regbyte & DS_STATUS_RST == 0:
            print("182", "error:   DS_STATUS_RST = 0")
            return(ERROR_YES)
        return(ERROR_NO)

    def DS_init(self,configbyte = 0xE1):
        """
        DS2428 Initialisation routine that sets the I2C address and then performs a
        device reset followed by writing the configuration byte.
        Parameter
            configbyte      any valid configuration byte
                0xF40   default,    external pull-up
                0xE1                DS internal pull-up
        Return
            1   init ok
            -1  init failed
        """

        if self.DS_reset() != ERROR_NO:  return(ERROR_YES)

        if self.DS_writeconfig(configbyte) != ERROR_NO: return(ERROR_YES)
        return(ERROR_NO)

    def DS_setchannel(self,channel):
        """
        Only DS2482-800.
        Selects, reads back and verifies the selected DS2482-800 channel
        Default is channel I00.

        Parameter
            channel 0, 1, 2, 3
        Return
            0..3 channel selected and verified
            -1   selection or verifying failed
        """
        if self.channel_last == channel:
            pass
        else:
            self.channel_last = channel
            # reset 1-wire
            if self.OW_reset() != ERROR_NO: return(ERROR_YES)

            # select channel
            if self.I2C_write_byte(cmd=DS_CMD_CHANNELSELECT,data=DS_CH_CODE[channel][0]) != ERROR_NO:
                return(ERROR_YES)

            if self.OW_busy() != 0:  return(ERROR_YES)
            regbyte = self.I2C_readbyte(regpointer = DS_REG_CHANNEL)
            if DS_CH_CODE[channel][1] != regbyte:
                print("DS_setchannel():\terror -->\tChannel selection or verifying failed")
                return(ERROR_YES)

        #print("Channel",channel)    #,hex(regbyte))
        return(channel)


    def DS_setreadpointer(self,regpointer):
        """
        Set Read Pointer
        Sets the read pointer to the specified register. Overwrites the read
        pointer position of any 1-Wire communication command in progress.
        Command Code: 0xE1,  Command parameter:  Pointer Codes

        Parameter
            regpointer  register to read
        Return
            1   set pointer ok
            -1  set pointer failed
        """

        if self.I2C_write_byte (cmd=DS_CMD_SETREADPOINTER,data=regpointer) != ERROR_NO:
            return(ERROR_YES)
        return(ERROR_NO)

    def DS_readstatus(self,statusbit=None):
        """
        Checks if a specific bit in the status register is set.
        Parameter
            statusbit
                a)  DS_STATUS_xx bit mask   or
                b)  None
        Return
            ds_stat
                0 or 1              case a)
                Statusbyte (8bit)   case b)
        """
        regbyte = self.I2C_readbyte(regpointer = DS_REG_STATUS)

        ds_stat = 0
        if statusbit == DS_STATUS_DIR:
            if regbyte & DS_STATUS_DIR  != 0:    ds_stat = 1
        elif statusbit == DS_STATUS_TSB:
            if regbyte & DS_STATUS_TSB  != 0:    ds_stat = 1
        elif statusbit == DS_STATUS_SBR:
            if regbyte & DS_STATUS_SBR  != 0:    ds_stat = 1
        elif statusbit == DS_STATUS_RST:
            if regbyte & DS_STATUS_RST  != 0:    ds_stat = 1
        elif statusbit == DS_STATUS_LL:
            if regbyte & DS_STATUS_LL   != 0:    ds_stat = 1
        elif statusbit == DS_STATUS_SD:
            if regbyte & DS_STATUS_SD   != 0:    ds_stat = 1
        elif statusbit == DS_STATUS_PPD:
            if regbyte & DS_STATUS_PPD  != 0:    ds_stat = 1
        elif statusbit == DS_STATUS_1WB:
            if regbyte & DS_STATUS_1WB  != 0:    ds_stat = 1
        else:
            ds_stat = regbyte
        return(ds_stat)

    def DS_writeconfig(self,confbyte):
        """
        Writes a new configuration byte. The new settings take effect
        immediately.
        Note:
        When writing to the Configuration Register, the new data is accepted
        only if the upper nibble (bits 7 to 4) is the one’s complement
        of the lower nibble (bits 3 to 0).
        When read, the upper nibble is always 0h.

        Command Code:       0xD2
        Command parameter:  Configuration Byte
        confbyte:   0xE1 = 1110 0001  ->  active DS 1-wire pull-up
                    0xF0 = 1111 0000  ->  external 1-wire pull-up resistor
        Return
            1   write ok
            -1  write failed
        """

        #print("Config byte: ","{0:>08b}".format(confbyte))
        if self.I2C_write_byte (cmd=DS_CMD_WRITECONFIG,data=confbyte) != ERROR_NO:
            return(ERROR_YES)
        regbyte = self.I2C_readbyte(regpointer=DS_REG_CONFIG)

        # when read, the upper nibble is always 0000.
        conf_lowernibble = confbyte & 0x0F
        if regbyte != conf_lowernibble:
            print("255", "error:    read low nibble conf byte","{0:>04b}".format(regbyte))
            return(ERROR_YES)
        if self.DS_readstatus(DS_STATUS_RST) != 0:
            print("error:  DS_STATUS_RST = 1")
            return(ERROR_YES)

        return(ERROR_NO)

    def OW_reset(self):
        """
        1-Wire Reset
        Generates a 1-Wire reset/presence-detect cycle at the 1-Wire line.
        The result is reported to the host processor through the
        Status Register, bits PPD and SD.Command Code: 0xB4

        Return
            1   reset ok
            -1  reset failed
        """

        if self.I2C_write_byte(OW_CMD_RESET) != ERROR_NO: return(ERROR_YES)
        if self.OW_busy() != 0: return(ERROR_YES)
        if self.DS_setreadpointer(DS_REG_STATUS) != ERROR_NO: return(ERROR_YES)
        if self.OW_busy() != 0:  return(ERROR_YES)
        # due to MAXIM APPLICATION NOTE 3684, used for ??
        if self.DS_readstatus(DS_STATUS_PPD) == 0:
            print("OW_reset() error  no PPD detected")
            return(ERROR_YES)
        time.sleep(OW_BUS_DELAY)
        return(ERROR_NO)

    # 1-wire Operations -----------------------------------------------

    def OW_writebit(self,vbit):
        """
        !!  Function not used and not tested

        1-Wire Single Bit
        Generates a single 1-Wire time slot with a bit value “V” as
        specified by the bit byte at the 1-Wire line. A "V" value of 0b
        generates a write-zero time slot; a "V" value of 1b generates a
        write-one time slot, which also functions as a read-data time slot

        Command Code:       0x87
        Commandparameter:  Bit_Byte

        Bit Allocation in the Bit_Byte          (x = don't care)
        BIT 7   BIT 6   BIT 5   BIT 4   BIT 3   BIT 2   BIT 1   BIT 0
        V       x       x       x       x       x       x       x

        Parameter
            1       write one-bit
            0       write zero-bit

        Return
            1   write sucessful
            -1  write failed
        """

        # set "V" , Bit7
        vbitbyte = 0x7F                         # V = 0
        if vbit == 1:    vbitbyte = 0xFF        # V = 1
        if self.I2C_write_byte (cmd=OW_CMD_SINGLEBIT, data=vbitbyte) != ERROR_NO:
            return(ERROR_YES)
        time.sleep(OW_BUS_DELAY)
        return(ERROR_NO)

    def OW_writebyte(self,writebyte):
        """
        Writes a single data byte to the 1-Wire line.
        Command Code: 0xA5 , Command parameter: Data Byte
        Parameter
            sendbyte    byte to write
        Return
            1   write ok
            -1  write failed
        """
        if self.I2C_write_byte (cmd=OW_CMD_WRITEBYTE, data=writebyte) != ERROR_NO:
            return(ERROR_YES)
        time.sleep(OW_BUS_DELAY)
        return(ERROR_NO)

    def OW_readbyte(self):
        """
        1-Wire Read Byte
        Generates eight read-data time slots on the 1-Wire line and stores
        the result in the Read Data Register. Command Code: 0x96

        Return
            1   read ok
            -1 read failed
        """

        if self.I2C_write_byte(cmd=OW_CMD_READBYTE) != ERROR_NO: return(ERROR_YES)
        time.sleep(OW_BUS_DELAY)
        return(ERROR_NO)

    def OW_writetriplet(self,dirbyte):
        """
        1-Wire Triplet
        Generates three time slots: two read time slots and one write time
        slot at the 1-Wire line. The type of write time slot depends on the
        result of the read time slots and the direction byte.
        The direction byte determines the type of write time slot if both
        read time slots are 0 (a typical case).
        In this case, the DS2482-100 generates a write-one time slot
        if V = 1 and a write-zero time slot if V = 0.

        If the read time slots are 0 and 1, they are followed by a
        write-zero time slot.
        If the read time slots are 1 and 0, they are followed by a
        write-one time slot.
        If the read time slots are both 1 (error case), the subsequent
        write time slot is a write-one.


        Command Code:       0x78
        Command parameter:  Direction Byte

        Bit Allocation in the Direction_Byte        (x = don't care)
        BIT 7   BIT 6   BIT 5   BIT 4   BIT 3   BIT 2   BIT 1   BIT 0
        V       x       x       x       x       x       x       x
        """

        if self.I2C_write_byte (cmd=OW_CMD_TRIPLET, data=dirbyte) != ERROR_NO:
            return(ERROR_YES)

        time.sleep(OW_BUS_DELAY)
        return(ERROR_NO)

    def OW_readrom(self):
        """
        The Read ROM [33h] command allows the bus master to read the
        8-bit family code, unique 48-bit serialnumber, and 8-bit CRC.
        This command should only be used if there is a single slave device
        on the bus.

        Return
            devrom  8 byte device rom
            -1       search rom failed
        """
        devrom = []

        # every master command sequence begins with a 1-wire reset
        if self.OW_reset() != ERROR_NO:
            print("OW_readrom():\terror -->\tOW_reset()")
            return(ERROR_YES)
        if self.OW_writebyte(OW_CMD_READROM) != ERROR_NO:
            print("OW_readrom():\terror -->\tOW_writebyte(OW_CMD_READROM))")
            return(ERROR_YES)

        # read 64 bit ( 8 byte) rom-code
        for i in range(0,8):
            if self.OW_busy() != 0:
                print("OW_readrom()\tOW_busy()failed")
                return(ERROR_YES)
            if self.OW_readbyte() != ERROR_NO:
                print("OW_readrom():\terror -->\tOW_read()")
                return(ERROR_YES)
            regbyte = self.I2C_readbyte(regpointer = DS_REG_DATA)
            devrom.append(regbyte)

        print("ROM byte\tHex\tInt")
        for i,regbyte in enumerate(devrom):
            print("\t{}\t{:x}\t{}".format(i,regbyte,regbyte))

        if self.crc8_check(devrom) != ERROR_NO:
            print("OW_readrom()\tCRC8 check failed")
            return(ERROR_YES)

        return(devrom)

    def OW_busy(self):
        """
        Checks if the bus is busy.
        Reads the DS2482 status-byte,  bit 1WB
        Returns:
            1   bus is busy         1WB = 1
            0   bus is not busy     1WB = 0
            -1  error
        """
        looplimit = 100
        loopcount = 0
        statbit_1WB = 1
        busycheck = 1
        while statbit_1WB !=0  and loopcount < looplimit:
            if self.DS_setreadpointer(DS_REG_STATUS) != ERROR_NO: return(ERROR_YES)
            statbit_1WB = self.DS_readstatus(DS_STATUS_1WB)
            busycheck = 0
            if statbit_1WB != 0:
                time.sleep(OW_BUS_DELAY)
                loopcount += 1
                busycheck = 1
        if loopcount >= looplimit:
            busycheck = -1
            print("OW_busy() check  failed")
        return(busycheck)


    def OW_singledev_info(self):
        """
        ! Make sure that only a single device is connected to the 1-wire bus.
        Otherwise the test will fail.

        Reads the devices 64 bit ROM-ID and prints the ID as an 8byte sequence.
        as hex and integer format.
        Starts with the last significant byte (LSB) which represents the
        family code)

        Calls the device with the  SKIP_ROM command.
        All 1-wire devices respond to this command but luckily there is
        no confusion. We have only one device on the bus which answers.

        Return
            1        Test ok
            -1       Test failed
        """
        devrom = self.OW_readrom()
        if devrom == -1:
            print("No device detected")
            return(ERROR_YES)

        famcode = devrom[0]
        print("Family Code\t",hex(famcode))
        print("Device ROM\t",end="")
        print(' '.join(map(hex, devrom)))

        return(ERROR_NO)

    def OW_detectdevice(self,devrom):
        """
        Resets the 1-wire bus and issues the SEARCH_ROM (0xF0) command with
        the desired 64bit (8byte) device rom.
        Each of the given 64 devrom bits should match the detected device bits.
        A difference finishes the loop and returns a "device not detected"
        The devrom bits are used as direction bits

        Parameter
            devrom:      64 bit device specific rom address ( CRC8,ID,Family code)

        Return
            1               device detected
            0               device not detected
            -1              detection error
        """

        detected = 1
        if self.OW_reset() != ERROR_NO: return(ERROR_YES)

        if self.OW_writebyte(OW_CMD_SEARCHROM) != ERROR_NO: return(ERROR_YES)
        # move through the 64 devrom bits
        bitselect = 0x1
        for byte in devrom:
            for i in range(0,8):
                # get next direction bit from the 64 bit devrom
                if byte & bitselect << i != 0:
                    dirbyte = 0x80
                    romdir = 1      # direction bit
                else:
                    dirbyte = 0x00
                    romdir = 0      # direction bit

                # get direction bit from the device
                if self.OW_writetriplet(dirbyte) != ERROR_NO: return(ERROR_YES)
                if self.OW_busy() != 0: return(ERROR_YES)

                romdetect = 0
                if self.DS_readstatus(DS_STATUS_DIR) != 0:
                    romdetect = 1

                if romdir != romdetect:
                    detected = 0    # difference, stop the loop
                    break
            if detected == 0: break
        return(detected)

    def crc8_check(self,data):
        """
        Compute the CRC8 byte of data and check with
        the dataCRC8  (MSB of data)

        Parameter
            data        the data of which CRC8 has to be calculated
            data_crc8   MSB of data, containes the CRC8
        Return
            1           calculated CRC8 = dataCRC8
            -1          verify of CRC8 failed
        """
        crc8 = 0
        for i in range(len(data)-1):    # exclude CRC8 byte
            byte = data[i]
            for j in range(8):
                fb_bit = (crc8 ^ byte) & 0x01
                if fb_bit == 0x01:
                    crc8 ^= 0x18
                crc8 = (crc8 >> 1) & 0x7f
                if fb_bit == 0x01:
                    crc8 |= 0x80
                byte >>= 1

        data_crc8 = data[len(data)-1]   # MSB, CRC8 byte only
        if crc8 != data_crc8:   return(ERROR_YES)
        return(ERROR_NO)

    def devcontrol(self, devrom, devtype,**kwargs):
        """
        Parameter:
            famcode_detected    detected device family code
            devrom      64 bit 1-wire device rom as hex string
            devtype

        Return:
            devdata     a 3 element tuple with specific device data
                [0]     general error field
                    -1  error
                    x   device specific or don't care
                [1]     float value, e.g Temperature
                [2]     string, e.g unit, error text,....
        """
        devdata = DEVDATA.copy()       # default
        famcode = devrom[0]

        if devtype == "DS18B20":
            devdata = self.ds18b20(devrom)

        elif devtype == "DS24SN":
            devdata = self.ds24sn(devrom)
        elif devtype == "DS2406":
            devdata = self.ds2406(devrom, ds2406_mode=kwargs["ds2406_mode"],ds2406_pio=kwargs["ds2406_pio"])
        else:
            if famcode == 0xFF:
                devdata[2] = "No 1-wire device detected"
                return(devdata)
            elif famcode == 0x0:
                devdata[2] = ("Given 1-wire device not found",devrom)
            else:
                devdata[2] = ("Unknown 1-wire device type",devtype)
                return(devdata)

        return(devdata)

    def ds18b20(self,devrom):
        """
        Provides the sensors temperature

        Parameter:
            devrom  64 bit device address
        Return
            devdata     a 3 element tuple with specific device data
                [0]     general error field
                    -1  error
                    x   device specific or don't care
                [1]     degrees celsius, -100 = error
                [2]     string, e.g unit, error text,....
         """
        DEVCMD_CONVERT_T        = 0x44 # measure the current temperature
        DEVCMD_READ_SCRATCHPAD  = 0xBE # master wants to read the scratchpad

        devdata = DEVDATA.copy()   # initialize the return list
        devdata[0] = -1     # general error
        devdata[1] = -1 #devrom     # just a value
        devdata[2] = "DS18B20 data error"     # general error

        scratchpad = []
        # every master command sequence begins with a 1-wire reset
        if self.OW_reset() != ERROR_NO: return(devdata)
        # to all:   listen
        if self.OW_writebyte(OW_CMD_SKIPROM) != ERROR_NO: return(devdata)
        # to all:   measure temperature
        if self.OW_writebyte(DEVCMD_CONVERT_T) != ERROR_NO: return(devdata)
        time.sleep(2)       # measuring needs some time

        if self.OW_reset() != ERROR_NO: return(devdata)
        # selects the device by calling its unique rom
        if self.OW_writebyte(OW_CMD_MATCHROM) != ERROR_NO:
            devdata[2] = "DS18B20 not detected"
            return(devdata)
        for byte in devrom:
            if self.OW_writebyte(byte) != ERROR_NO: return(devdata)
            time.sleep(OW_BUS_DELAY)
        # read the memory map
        if self.OW_writebyte(DEVCMD_READ_SCRATCHPAD) != ERROR_NO: return(devdata)
        # Figure 9. DS18B20 Memory Map
        for i in range(0,9):
            if self.OW_busy() != 0: return(devdata)
            if self.OW_readbyte() != ERROR_NO: return(devdata)
            regbyte = self.I2C_readbyte(regpointer = DS_REG_DATA)
            scratchpad.append(regbyte)

        # verify memory read with CRC8
        if self.crc8_check(scratchpad) != ERROR_NO:
            devdata[2] = "CRC8 check failed"
            return(devdata)

        # combine memory map byte 0 and byte 1 to a 16 bit word
        therm_raw =scratchpad[1] << 8 | scratchpad[0]

        # check for negative temperature
        if therm_raw & 0x8000 != 0:             # test most significant bit
            #2's complement = value - 2**length = 2**16 = 0x10000
            therm_raw -= 0x10000  # 2's complement

        # read thermometer resolution. Bits 5 and 6 in the Configuration Register
        therm_resolution = scratchpad[4] & 0x60
        if therm_resolution == 0xC0:
            pass
        elif therm_resolution == 0x40:
            therm_raw <<= 1
        elif therm_resolution == 0x20:
            therm_raw <<= 2
        elif therm_resolution == 0x00:
            therm_raw <<= 3
        else:
            pass
        celsius = therm_raw / 16.0

        devdata[0] = 1
        devdata[1] = celsius
        devdata[2] = "°C"
        return(devdata)


    def ds24sn(self,devrom):
        """
        Ds2401, DS2411 are Silicon Number devices and often used as on/off detection.
        If the device is not detected on the 1-wire bus it's assumed
        to be switched off.
        If the device is detected it's assumed to be switched on.
        """
        devdata = DEVDATA.copy()   # initialize the return list
        devdata[2] = "DS24SN error"

        detected = self.OW_detectdevice(devrom)
        if detected == ERROR_YES: return(devdata)
        devdata[0] = 1
        if detected == 0:
            devdata[1] = 0
            devdata[2] = "DS24SN not detected"
        else:
            devdata[1] = 1
            devdata[2] = "DS24SN detected"
        return(devdata)

    # noinspection PyUnresolvedReferences
    def ds2406(self, devrom, ds2406_mode="w1", ds2406_pio='A'):
        """
        Allows to use the DS2406 PIO pin in write or read mode.
        Parameter:
            devrom  64 bit device address
            mode
                w0  write, use PIO as output, set output high
                w1  write, use PIO as output, set output low
                r   read, use PIO as input
            channel
                A   PIO A
                B   PIO B
                AB  both    not implemented

        Return
           devdata     a 3 element tuple with specific device data
                [0]     general error field
                    0   no error
                    -1  error
                [1]     channel info byte
                [2]     selected channel A / B
                        or error info text

        ############
        Table 1. Channel Control Byte 1
        BIT 7   BIT 6   BIT 5   BIT 4   BIT 3   BIT 2   BIT 1   BIT 0
        ALR     IM      TOG     IC      CHS1    CHS0    CRC1    CRC0

        Case 1:
        only channel A, only as output, CRC disabled, edge detection not changed:
        0       0       0       0       0       1       0       0   =0x4

        Case 2:
        only channel A, only as input, CRC disabled, edge detection not changed:
        0       0       1       0       0       1       0       0   =0x24

        CHS1    CHS0    Description
        0       0       (not allowed)
        0       1       channel A only
        1       0       channel B only
        1       1       both channels interleaved

        CRC1 CRC0    Description
        0    0       CRC disabled (no CRC at all)
        0    1       CRC after every byte
        1    0       CRC after 8 bytes
        1    1       CRC after 32 bytes

        TOG IM  Chan    Effect
        0   0   one     Write all bits to the selected channel
        0   1   one     Read all bits from the selected channel
        1   0   one     Write 8 bits, read 8 bits, write, read, etc. to/from
        1   1   one     Read 8 bits, write 8 bits, read, write, etc. from/to
        0   0   two     Repeat: four times (write A, write B)
        0   1   two     Repeat: four times (read A, read B)
        1   0   two     Four times: (write A, write B), four times: (read A, read B), write, read, etc.
        1   1   two     Four times: (read A, read B), four times: (write A,write B), read, write, etc.

        CHANNEL INFO BYTE Figure 9
        BIT 7   Supply Indication       0 = no supply
        BIT 6   Number of  Channels     0 = channel A only
        BIT 5   PIO-B   Activity Latch
        BIT 4   PIO-A   Activity Latch
        BIT 3   PIO B   Sensed  Level
        BIT 2   PIO A   Sensed Level
        BIT 1   PIO-B   Channel Flip-Flop Q
        BIT 0   PIO-A   Channel Flip-Flop Q

        """
        DEVCMD_CHANNELACCESS    = 0xF5  # access the PIO channel

        devdata = DEVDATA.copy()   # initialize the return list
        devdata[0] = -1     # general error
        devdata[1] = -1     # just a value
        devdata[2] = "DS2406 data error"

        detected = self.OW_detectdevice(devrom)
        if detected == ERROR_YES: return(devdata)
        if detected == 0:
            devdata[1] = 0
            devdata[2] = "DS24SN not detected"
            return(devdata)

        if self.OW_reset() != ERROR_NO: return(devdata)

        # selects the device by calling its unique rom
        if self.OW_writebyte(OW_CMD_MATCHROM) != ERROR_NO: return(devdata)
        for byte in devrom:
            if self.OW_writebyte(byte) != ERROR_NO: return(devdata)
            time.sleep(OW_BUS_DELAY)

        # set the channel control byte
        if self.OW_writebyte(DEVCMD_CHANNELACCESS) != ERROR_NO: return(devdata)

        # Use the write bit command to send the desired output state.
        # You can change the output state as many times as
        # necessary, until a 1-Wire reset command is sent.

        # write channel control byte 1
        if self.OW_writebyte(0x04) != ERROR_NO: return(devdata)
        # write channel control byte 2, always 0xFF
        if self.OW_writebyte(0xFF) != ERROR_NO: return(devdata)

        # read channel info byte
        if self.OW_busy() != 0: return(devdata)
        if self.OW_readbyte() != ERROR_NO: return(devdata)
        regbyte = self.I2C_readbyte(regpointer = DS_REG_DATA)
        pio_info = regbyte

        if ds2406_mode     == "w1":  writeread = 1 # set output high
        elif ds2406_mode   == "w0":  writeread = 0 # set output low
        elif ds2406_mode   == "r":
            devdata[2] = ("DS2406 read mode not implemented",ds2406_mode)
            return(devdata)
        else:
            devdata[2] = ("DS2406 mode not known", ds2406_mode)
            return(devdata)
        if self.OW_writebit(writeread) != ERROR_NO: return(devdata)

        devdata[0] = 1
        devdata[1] = pio_info
        devdata[2] = ds2406_pio
        #print("865",DEVDATA, devdata)
        return(devdata)



    def class_terminate(self):
        """
        """
        self.pi.i2c_close(self.dsh)
        return()

def app_init(pi,dsr,I2C_address, DS2482_address, OW_pullup):
    #def app_init(pi,cfgyaml,dsr):
    """
    initial settings
    """


    classDS = dsr.DS2482(pi,I2C_address, DS2482_address)

    #print("\n### Reset Ds2482 and set the configuration ------------")
    if classDS.DS_init(OW_pullup) == -1:
        print("war nix")
        pass

    return(classDS)

def app_close(classDS):
    """
    """
    classDS.class_terminate()
    return()

if __name__ == "__main__":

    exec(open("DS2482rsf_run.py").read())
